structure PeirceExp =
   struct
       datatype rexp = ConR of string
                     | PlusR of rexp list 
                     | ProdR of rexp * rexp
                     | UnitR
                     | CompR of rexp
                     | Id of sexp
                     | Prod of rexp * rexp
                     | Conv of rexp
                     | Cylin of sexp
       and      sexp = ConS of string 
                     | PlusS of sexp list
                     | ProdS of sexp * sexp
                     | UnitS 
                     | CompS of sexp
                     | PreIm of rexp * sexp   (* R : B *)

       type set = sexp 
       type rel = rexp

       val ZeroS = PlusS []
       fun SumS (A,B) = PlusS [A,B]

       val ZeroR = PlusR []
       fun SumR (S,T) = PlusR [S,T]
       val I = Id UnitS
   end

structure PeirceTerm = 
  struct
    open PeirceExp  (* fuegt Typen und Werte aus PeirceExp ein *)

    val Anna    = ConS "Anna"
    val Dieter  = ConS "Dieter"
    val Emil    = ConS "Emil"
    val Maria   = ConS "Maria"

    val Frau    = ConS "Frau"
    val Mann    = ConS "Mann"
    val Student = ConS "Student"

    val Bruder  = ConR "Bruder"
    val Nachbar = ConR "Nachbar"

    val arbeiten = ConS "arbeiten"

    val kennen = ConR "kennen"
    val lieben = ConR "lieben"

    val gesund = ConS "gesund"

    structure B = Boole
    fun E(A:set) = not (A = ZeroS)    (* Vorgriff auf Boettner *)

    fun toString (PlusS[])    = " 0 " 
      | toString UnitS        = " 1 " 
      | toString (ConS(w))    = w
      | toString (PlusS[A,B]) = 
        "(" ^ (toString A) ^ " + " ^ (toString B) ^ ")"
      | toString (ProdS(A,B)) = 
        "(" ^ (toString A) ^ " . " ^ (toString B) ^ ")"
      | toString (CompS(A))   = 
        " -" ^ (toString A) 
      | toString (PreIm(R,A)) = 
        "(" ^ (RtoString R) ^ " : " ^ (toString A) ^ ")"
      | toString _ = "noch nicht definiert"
    and RtoString (ConR(r))    = r
      | RtoString UnitR        = " 1' " 
      | RtoString (PlusR[])    = " 0' " 
      | RtoString (PlusR[R,S]) = 
        "(" ^ (RtoString R) ^ " + " ^ (RtoString S) ^ ")"
      | RtoString (ProdR(R,S)) = 
        "(" ^ (RtoString R) ^ " . " ^ (RtoString S) ^ ")"
      | RtoString (CompR(R))   = "-" ^ (RtoString R) 
      | RtoString (Prod(R,S))  = 
        "(" ^ (RtoString R) ^ " ; " ^ (RtoString S) ^ ")"
      | RtoString (Conv(R))    = "~" ^ (RtoString R) 
      | RtoString (Cylin(A))   = " 1x" ^ (toString A) 
      | RtoString _ = "noch nicht definiert"
  end

signature UNIVERSE = sig eqtype element 
                         val elements : element list 
                     end

signature PEIRCEfull = 
   sig 
       include PEIRCE
       type element 
       val ConS : element list -> set
       val ConR : (element * element) list -> rel
   end                     
    
functor Peirce(U : UNIVERSE) : PEIRCEfull =
   struct
     type element = U.element
     fun ConS(A:element list) = Set.mk A
     fun ConR(R:(element*element) list) = Set.mk R

     type set = element Set.set

     val UnitS = Set.mk U.elements
     val ZeroS = Set.empty
     fun SumS(A,B) = Set.union A B
     fun ProdS(A,B) = Set.intersection A B
     fun CompS(A) = Set.difference UnitS A
 
     type rel = (element * element) Set.set

     fun Id(A) = Set.map (fn x => (x,x)) A
     val I = Id(UnitS)
     val UnitR:rel = 
         let fun f x = List.map (fn y => (x,y)) U.elements
         in 
             Set.mk (List.foldr (op @) nil (List.map f U.elements))
         end
     val ZeroR = Set.empty
     fun SumR(R,S) = Set.union R S
     fun ProdR(R,S) = Set.intersection R S
     fun CompR(R) = Set.difference UnitR R
     fun Conv(R) = Set.map (fn (x,y) => (y,x)) R
     fun Prod(R,S) = R             (* noch definieren! *)
     fun PreIm(R,A) = A            (* noch definieren! *)
     fun Cylin(A) = UnitR          (* noch definieren! *)
   end 

structure PeirceFull = Peirce(datatype element = a | b | c | d | e  
                              val elements = [a,b,c,d,e])

local
    datatype element = a | b | c | d | e  
    structure P = Peirce(type element = element
                         val elements = [a,b,c,d,e])
in 
    structure Dext = struct 
      open P
      val s = ConS [a,b]
      val r = ConR [(a,b),(b,c)]
    end : sig 
              include PEIRCE 
              val s:set 
              val r:rel 
          end
end

